#include <mpi.h>
#include <iostream>
#include <stdio.h> //Files
#include <iomanip> //SetPresicion
#include <time.h>

#include <fstream>

MPI_Status status;

using namespace std;

const int NV=60;
const int NE=80;
const int Max_Number_of_Cores=1000;
const int Number_of_Edges_to_Graph_Send=30;
const int Number_of_Recursions_to_Ask_Help=50;

int Number_of_Recursions=0;

int tag0=1;
int my_rank;
double help_me=5;


struct Graph
	{ int KAO[NV+1];
      int FO[2*NE];
      double PArray[2*NE];
      int VertNumb;
	  int EdgNumb;
	};

// serial functions

Graph GraphFromFile();
void ResultOutput(double res, time_t Tbeg, time_t Tend);
double FactoringProbability(Graph G);

// MPI functions

double Factoring_MPI_MasterProcess(Graph G, int Number_of_Processes);
void Factoring_MPI_HostProcess();


int main(int argc, char **argv)
{
 MPI_Init(&argc, &argv);
 int rank, size;
 MPI_Comm_rank(MPI_COMM_WORLD, &rank);
 MPI_Comm_size(MPI_COMM_WORLD, &size);
 
 if (size==1)
 {  time_t Tbeg,Tend;
	Graph G=GraphFromFile();
    double Result=0;
	time(&Tbeg);
	Result=FactoringProbability(G);
	time(&Tend);
    ResultOutput(Result,Tbeg,Tend);
 } else
 if (rank == 0)
  { time_t Tbeg,Tend;
	Graph G=GraphFromFile();
    double Result=0;
	
	// for mili sec precision
	//std::cout << std::fixed << std::setprecision(25);
	//double startTime = MPI_Wtime(); 
	time(&Tbeg);
		
	Result=Factoring_MPI_MasterProcess(G,size);
	
	//double endTime = MPI_Wtime();	
	//std::cout << "working time(ms) = " << endTime - startTime << "\n";

	time(&Tend);
    ResultOutput(Result,Tbeg,Tend);
  }
 else
 {   my_rank=rank;
	 Factoring_MPI_HostProcess();
 }
    
  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Finalize();

 return 0;
}

Graph GraphFromFile()
{	char ch;	Graph G;	int i;
	FILE *F; 
    F = fopen("c:\\ \\Programs\\Graphs\\grid67.txt","rt");
    fscanf( F, "%d", &G.VertNumb ); 
    fscanf( F, "%d", &G.EdgNumb );
 	for (i=0; i<G.VertNumb; i++)
    {fscanf( F, "%d", &G.KAO[i] );
	 fscanf( F, "%c", &ch );}
     fscanf( F, "%d", &G.KAO[G.VertNumb] );
	for (i=0; i<G.EdgNumb*2-1; i++)
    {fscanf( F, "%d", &G.FO[i] );
	 fscanf( F, "%c", &ch );	}
    fscanf( F, "%d", &G.FO[G.EdgNumb*2-1] );
	fclose(F);
	for (int i=0; i<2*G.EdgNumb; i++) G.PArray[i]=0.9;
	return G;
}

void ResultOutput(double res, time_t Tbeg, time_t Tend)
{ 	 int WorkTime;

     WorkTime=int(Tend)-int(Tbeg);
	 
	 std::cout << std::fixed << std::setprecision(20);
	 cout << "working time(sec) = " << WorkTime << "\n";
	 cout << "Result = " << res << "\n";
	
     FILE *F; 

	 F = fopen("result.txt","wt");
	 fprintf( F, "res: %.20f", res );
	 fprintf( F, "  time: %d", WorkTime ); 
	 fclose(F);
}

bool Connectivity(Graph G)
{	int A[NV+1], B[NV+1], Spot[NV+1];
	int LB, i, j, sum=1, l=1;
	
	A[0]=1;
	for (i=0; i<1+G.VertNumb; i++) Spot[i]=0;
	Spot[1]=1;
	while (l>0)
	{   LB=0;
		for (i=0; i<l; i++)
		{   for (j=G.KAO[A[i]-1]; j<G.KAO[A[i]]; j++)
			{   if (Spot[G.FO[j]]==0)
				{	LB++;
					B[LB-1]=G.FO[j];
					Spot[G.FO[j]]=1;
					sum++;
				}
			}
		}
        l=LB;
		if (l>0)
		{	for (i=0; i<l; i++) A[i]=B[i];	}
	}
  if (sum==G.VertNumb)  {return true;} else {return false;}
  if (G.VertNumb==0)	{return true;}
}

int SearchEdge(Graph G, int i, int j)
{	int  k;	bool b=true;
	for (k=G.KAO[i-1]; k<G.KAO[i]; k++)
	{	if (G.FO[k]==j) {   b=false;
			                return k;}
	}
	if (b==true) {return G.EdgNumb*2;}
}

Graph DeleteEdge(Graph G, int u, int v)
{	int i, e, f;
	Graph R;
	//e=SearchEdge(G,u,v);
	bool b=true;
	for (int k=G.KAO[u-1]; k<G.KAO[u]; k++)
	{if (G.FO[k]==v) {   b=false;
			                e=k;}
	}if (b==true) {e=G.EdgNumb*2;}
 //   f=SearchEdge(G,v,u);
    b=true;
	for (int k=G.KAO[v-1]; k<G.KAO[v]; k++)
	{if (G.FO[k]==u) {   b=false;
			                f=k;}
	}if (b==true) {f=G.EdgNumb*2;}
	//
	if ((e<G.EdgNumb*2) && (f<G.EdgNumb*2)) 
	{  	R.VertNumb=G.VertNumb;
		R.EdgNumb=G.EdgNumb-1;
		if (e<f) 
		{ for (i=0; i<u; i++) {R.KAO[i]=G.KAO[i];}
		  for (i=u; i<v; i++) {R.KAO[i]=G.KAO[i]-1;}
          for (i=v; i<=G.VertNumb; i++) {R.KAO[i]=G.KAO[i]-2;}
		  if (e>0)
		  {for (i=0; i<e; i++)
			  { R.FO[i]=G.FO[i];
			    R.PArray[i]=G.PArray[i];}
		  }
		  if (f>e+1)
		  {for (i=e; i<f-1; i++)
			  { R.FO[i]=G.FO[i+1];
			    R.PArray[i]=G.PArray[i+1];}
		  }
		  if (G.EdgNumb*2>f+1)
		  {for (i=f-1; i<G.EdgNumb*2-2; i++)
			  { R.FO[i]=G.FO[i+2];
	            R.PArray[i]=G.PArray[i+2];}
		  }
		}else
		{ for (i=0; i<v; i++) {R.KAO[i]=G.KAO[i];}
		  for (i=v; i<u; i++) {R.KAO[i]=G.KAO[i]-1;}
          for (i=u; i<=G.VertNumb; i++) {R.KAO[i]=G.KAO[i]-2;}
		  if (f>0)
		  {for (i=0; i<f; i++)
			  { R.FO[i]=G.FO[i];
			    R.PArray[i]=G.PArray[i];}
		  }
		  if (e>f+1)
		  {for (i=f; i<e-1; i++)
			  { R.FO[i]=G.FO[i+1];
			    R.PArray[i]=G.PArray[i+1];}
		  }
		  if (G.EdgNumb*2>e+1)
		  {for (i=e-1; i<G.EdgNumb*2-2; i++)
			  { R.FO[i]=G.FO[i+2];
		        R.PArray[i]=G.PArray[i+2];}
		  }
		}
	}
	return R;
}

Graph MergeVertex(Graph G, int cutv1, int cutv2)
{   int KAO[NV+1], FO[2*NE], S[NV+1];
	double PArray[2*NE];
    int i, j, l=0, k, boolka;

	for (i=1; i<cutv2; i++) S[i]=i;
	S[cutv2]=cutv1;
    for (i=cutv2+1; i<G.VertNumb+1; i++) S[i]=i-1;
	KAO[0]=0;
	for (i=1; i<G.VertNumb+1; i++)
	{ if (i==cutv1) {boolka=1;} else
      {if (i==cutv2) {boolka=2;} else
       { k=0;
         for (j=G.KAO[i-1]; j<G.KAO[i]; j++)
	     {if ((G.FO[j]==cutv1) || (G.FO[j]==cutv2)) {k++;} }
          if (k==2) {boolka=3;} else {boolka=4;}
       }
      }
	switch (boolka)
   	{
	case 1:
		KAO[S[i]]=KAO[S[i]-1];
        for (j=G.KAO[i-1]; j<G.KAO[i]; j++)
        {if (G.FO[j]!=cutv2)
		{++KAO[S[i]];
		 FO[l]=S[G.FO[j]];
		// k=SearchEdge(G,cutv2,G.FO[j]);
		 bool b=true;
	     for (int m=G.KAO[cutv2-1]; m<G.KAO[cutv2]; m++)
         {if (G.FO[m]==G.FO[j]) {   b=false;
	        	                    k=m;}
	     }if (b==true) {k=G.EdgNumb*2;}
	//
		 if (k!=G.EdgNumb*2)
		 {PArray[l]=G.PArray[j]+G.PArray[k]-G.PArray[j]*G.PArray[k];}
		 else {PArray[l]=G.PArray[j];}
		 ++l;
		}
		}
        for (j=G.KAO[cutv2-1]; j<G.KAO[cutv2]; j++)
        {if (G.FO[j]!=cutv1)
		{//if (SearchEdge(G,cutv1,G.FO[j])==G.EdgNumb*2)
          bool b=true;
		  int o;
	      for (int m=G.KAO[cutv1-1]; m<G.KAO[cutv1]; m++)
          {if (G.FO[m]==G.FO[j]) {   b=false;
	        	                     o=m;}
	      }if (b==true) {o=G.EdgNumb*2;} if (o==G.EdgNumb*2)
		//
		  {++KAO[S[i]];
		 FO[l]=S[G.FO[j]];
		 PArray[l]=G.PArray[j];
		 ++l;
		}
		}
		}
		break;
	case 2:
		break;
    case 3:
        KAO[S[i]]=KAO[S[i]-1]+G.KAO[i]-G.KAO[i-1]-1;
        for (j=G.KAO[i-1]; j<G.KAO[i]; j++)
        {if (G.FO[j]!=cutv2)
		{if (G.FO[j]!=cutv1) 
		{FO[l]=S[G.FO[j]];
         PArray[l]=G.PArray[j];
         ++l;}
		else
		{
		  FO[l]=S[G.FO[j]];
          for (k=G.KAO[i-1]; k<G.KAO[i]; k++)
          {
		  if (G.FO[k]==cutv2) 
          {PArray[l]=G.PArray[j]+G.PArray[k]-G.PArray[j]*G.PArray[k];}
          }
		 ++l;
		}
		}
		}
		break;
    case 4:
        KAO[S[i]]=KAO[S[i]-1]+G.KAO[i]-G.KAO[i-1];
        for (j=G.KAO[i-1]; j<G.KAO[i]; j++)
		{
         FO[l]=S[G.FO[j]];
         PArray[l]=G.PArray[j];
         ++l;
		}
		break;
	}}

	G.VertNumb--;
	G.EdgNumb=l/2;
    for (i=0; i<=G.VertNumb; i++) G.KAO[i]=KAO[i];
    for (i=0; i<G.EdgNumb*2; i++) {G.FO[i]=FO[i];
	                               G.PArray[i]=PArray[i];}
	return G;
}

Graph ShootsDeletion(Graph G, double &p)
{ Graph H;
  int i,j,u,v;
  bool b;
  int Deg[NV+1], Numbers[NV+1];

  p=1;
  for (i=1; i<=G.VertNumb; i++) Deg[i]=G.KAO[i]-G.KAO[i-1];
  for (i=1; i<=G.VertNumb; i++)
  {	  b=true;
	  v=i;
	  while (b)
	  { if (Deg[v]==1)
	    {for (j=G.KAO[v-1]; j<G.KAO[v]; j++)
	      {if (Deg[G.FO[j]]>0)
	          {p=p*G.PArray[j]; 
	           u=G.FO[j];
			   break;
	          }
	      }
	     Deg[v]=0;
		 Deg[u]--;
		 v=u;   
	    } else {b=false;}
	  }

  }
  u=0;
  H.KAO[0]=0;
  for (i=1; i<=G.VertNumb; i++)
  {if (Deg[i]>0)
      {u++;
       H.KAO[u]=H.KAO[u-1]+Deg[i];
	   Numbers[i]=u;
      }
  }
  if (u>0)
  {H.VertNumb=u;
   H.EdgNumb=H.KAO[u]/2;
   u=0;
   for (i=1; i<=G.VertNumb; i++)
   {if (Deg[i]>0) {for (j=G.KAO[i-1]; j<G.KAO[i]; j++) {if (Deg[G.FO[j]]>0){
                                                           H.FO[u]=Numbers[G.FO[j]];
                                                           H.PArray[u]=G.PArray[j];
                                                           u++;}
                                                        }
  
  
                  }
   }
  } else // graph with only one vertice has been obtained
  {H.VertNumb=1;
   H.EdgNumb=0;
  }
  return H;
}

double ProbabilityOf5VertexGraph(Graph G)
{int u,v,SE;

 if(G.VertNumb==5)
 {double a,b,c,d,e,f,q,h,k,l,k1,k2,k3,k4,k5,k6,k7,k8,k9,k10,k11,k12,k13,k14,k15;
  double R[4][4];
  for (u=1; u<5; u++)
  {for (v=u+1; v<6; v++)
   {bool bol=true;
	for (int k=G.KAO[u-1]; k<G.KAO[u]; k++)
	{if (G.FO[k]==v) {   bol=false;
			                SE=k;}
	}if (bol==true) {SE=G.EdgNumb*2;}
	if (SE<G.EdgNumb*2)
    {R[u-1][v-2]=1-G.PArray[SE];}
	else R[u-1][v-2]=1;
   }
  }
  a=R[0][0]; b=R[0][1]; c=R[0][2]; d=R[0][3]; e=R[1][1]; 
  f=R[1][2]; q=R[1][3]; h=R[2][2]; k=R[2][3]; l=R[3][3];
  k1=1-e*(f*q+h*k);
  k2=1-h*(b*k+c*l);
  k3=1-l*(c*f+d*q);
  k4=1-d*(a*b+q*k);
  k5=1-a*(b*c+e*f);
  k6=(1-a)*(1-h)*(1-k)+(1-a)*(1-l)*((1-h)*k+h*(1-k))+a*h*k*(1-4*l);
  k7=(1-c)*(1-d)*(1-e)+(1-e)*(1-l)*((1-c)*d+c*(1-d))+d*e*l;
  k8=(1-a)*(1-d)*(1-h)+(1-q)*(1-h)*((1-a)*d+a*(1-d))+a*h*q;
  k9=(1-a)*(1-b)*(1-l)+(1-e)*(1-l)*((1-a)*b+a*(1-b))+a*e*l;
  k10=(1-e)*(1-d)*(1-f)+(1-d)*(1-h)*((1-e)*f+e*(1-f))+d*f*h;
  k11=(1-b)*(1-f)*(1-q)+(1-b)*(1-l)*((1-f)*q+f*(1-q))+b*q*l;
  k12=(1-c)*(1-e)*(1-q)+(1-c)*(1-k)*((1-e)*q+e*(1-q))+c*e*k;
  k13=(1-b)*(1-d)*(1-f)+(1-f)*(1-k)*((1-b)*d+b*(1-d))+b*d*f;
  k14=(1-b)*(1-c)*(1-q)+(1-q)*(1-h)*((1-b)*c+b*(1-c))+b*c*q;
  k15=(1-a)*(1-c)*(1-k)+(1-f)*(1-k)*((1-a)*c+a*(1-c))+c*f*k;
   return 1-(b*c*(a*d*k1+f*e*(d*q*k6+k*l*k8))+
             f*q*(a*e*k2+h*k*(a*b*k7+c*d*k9))+
             b*h*(e*k*k3+d*l*(a*f*k12+e*q*k15))+
             c*l*(f*h*k4+a*q*(b*k*k10+e*h*k13))+
             d*k*(q*l*k5+a*e*(c*h*k11+f*l*k14))
			 );
  } else
	 if (G.VertNumb==4)
	 {double a,b,c,d,e,f;
	  double R[3][3];
      for (u=1; u<4; u++)
      {for (v=u+1; v<5; v++)
       {bool bol=true;
	    for (int k=G.KAO[u-1]; k<G.KAO[u]; k++)
	    {if (G.FO[k]==v) {   bol=false;
		    	                SE=k;}
 	    }if (bol==true) {SE=G.EdgNumb*2;}
	    if (SE<G.EdgNumb*2)
        {R[u-1][v-2]=1-G.PArray[SE];}
	    else R[u-1][v-2]=1;
       }
      }
	  a=R[0][0]; b=R[1][1]; c=R[2][2]; d=R[0][2]; e=R[1][2]; f=R[0][1];
	  return 1-(6*a*b*c*d*e*f-2*(b*d*e*f*(a+c-0.5)+a*c*e*f*(b+d-0.5)+a*b*c*d*(e+f-0.5))+a*b*e+a*d*f+b*c*f+c*d*e);
	 } else if (G.VertNumb==3)
	 {double R[2][2];
      for (u=1; u<3; u++)
      {for (v=u+1; v<4; v++)
       {bool bol=true;
	    for (int k=G.KAO[u-1]; k<G.KAO[u]; k++)
	    {if (G.FO[k]==v) {   bol=false;
		  	                SE=k;}
	    }if (bol==true) {SE=G.EdgNumb*2;}
	    if (SE<G.EdgNumb*2)
        {R[u-1][v-2]=G.PArray[SE];}
	     else R[u-1][v-2]=0;
	   }
	  }
      return (R[0][0]*R[0][1]*R[1][1]+(1-R[0][0])*R[0][1]*R[1][1]+R[0][0]*(1-R[0][1])*R[1][1]+R[0][0]*R[0][1]*(1-R[1][1]));
	 } else  if (G.VertNumb==2) {return G.PArray[0];} else {return 1;}
}

Graph ParallelSeriesTransformation(Graph G, double &p)
{bool b=false;
 int i,j,u,v,w,l,k,SE;
 double p1,p2,p3;
 
 p=1;
 for (k=1; k<=G.VertNumb; k++)
 {if (G.KAO[k]-G.KAO[k-1]==2)
  {b=true;
   v=k;
   u=G.FO[G.KAO[k-1]];
   w=G.FO[G.KAO[k]-1];
   p1=G.PArray[G.KAO[k-1]];
   p2=G.PArray[G.KAO[k]-1];
   break;
  }
 }
  while (b) 
 {int Numbers[NV+1];
  Graph H;
  p=p*(p1+p2-p1*p2);
  p3=p1*p2/(p1+p2-p1*p2);
  H.VertNumb=G.VertNumb-1;
  H.KAO[0]=0;
  for (i=1; i<=v-1; i++) Numbers[i]=i;
  Numbers[v]=0;
  for (i=v+1; i<=G.VertNumb; i++) Numbers[i]=i-1;
  l=0;
 
  bool bol=true;
  for (int m=G.KAO[u-1]; m<G.KAO[u]; m++)
	{if (G.FO[m]==w) {   bol=false;
			                SE=m;}
	}if (bol==true) {SE=G.EdgNumb*2;}
	
	if (SE<G.EdgNumb*2)
    {p3=1-(1-G.PArray[SE])*(1-p3);
     H.EdgNumb=G.EdgNumb-2;
     for (i=1; i<=G.VertNumb; i++) 
     {if ((i==u) || (i==w))
      {H.KAO[Numbers[i]]=H.KAO[Numbers[i]-1];
       for (j=G.KAO[i-1] ; j<=G.KAO[i]-1; j++) 
	   {if ((G.FO[j]==w) || (G.FO[j]==u))
	     {H.FO[l]=Numbers[G.FO[j]];
           H.PArray[l]=p3;
           H.KAO[Numbers[i]]++;
           l++;
         } else if (G.FO[j]!=v)
         {H.FO[l]=Numbers[G.FO[j]];
           H.PArray[l]=G.PArray[j];
           H.KAO[Numbers[i]]++;
           l++;
         }
	   }
     } else if (i!=v)
     {H.KAO[Numbers[i]]=H.KAO[Numbers[i]-1];
      for (j=G.KAO[i-1] ; j<=G.KAO[i]-1; j++) 
	  {H.FO[l]=Numbers[G.FO[j]];
       H.PArray[l]=G.PArray[j];
       H.KAO[Numbers[i]]++;
       l++;
      }
	 }
   }
  } else
  {H.EdgNumb=G.EdgNumb-1;
   for (i=1; i<=G.VertNumb; i++) 
   {if (i==u)
    {H.KAO[Numbers[i]]=H.KAO[Numbers[i]-1];
     H.FO[l]=Numbers[w];
     H.PArray[l]=p3;
     H.KAO[Numbers[i]]++;
     l++;
     for (j=G.KAO[i-1]; j<=G.KAO[i]-1; j++) 
	 {if (G.FO[j]!=v)
      {H.FO[l]=Numbers[G.FO[j]];
       H.PArray[l]=G.PArray[j];
       H.KAO[Numbers[i]]++;
       l++;
	  }
	 }
    } else if (i==w)
    {H.KAO[Numbers[i]]=H.KAO[Numbers[i]-1];
     H.FO[l]=Numbers[u];
     H.PArray[l]=p3;
     H.KAO[Numbers[i]]++;
     l++;
     for (j=G.KAO[i-1]; j<=G.KAO[i]-1; j++) 
	 {if (G.FO[j]!=v)
      {H.FO[l]=Numbers[G.FO[j]];
       H.PArray[l]=G.PArray[j];
       H.KAO[Numbers[i]]++;
       l++;
	  }
	 }
	} else if (i!=v)
	{H.KAO[Numbers[i]]=H.KAO[Numbers[i]-1];
     for (j=G.KAO[i-1]; j<=G.KAO[i]-1; j++) 
	 {H.FO[l]=Numbers[G.FO[j]];
      H.PArray[l]=G.PArray[j];
      H.KAO[Numbers[i]]++;
      l++;
     }
	}
   }
  }
  G=H;
  b=false;
  for (k=1; k<=G.VertNumb; k++)
  {if (G.KAO[k]-G.KAO[k-1]==2)
   {b=true;
    v=k;
    u=G.FO[G.KAO[k-1]];
    w=G.FO[G.KAO[k]-1];
    p1=G.PArray[G.KAO[k-1]];
    p2=G.PArray[G.KAO[k]-1];
    break;
   }
  }
 }
  return G;
}

double Branching(Graph G, int variant)
{   int u,v,vt,ut,deg,i;
	double p,pst_factor;
	
   // Number_of_Recursions++;
	if ((variant==0) && (Connectivity(G)==false)) {return 0;}
	else
	{G=ParallelSeriesTransformation(G,pst_factor);
     if (G.VertNumb>5)
	 {  vt=G.VertNumb;
	    deg=2*G.VertNumb+2;
		for (i=G.EdgNumb*2-1;  i>=G.KAO[1]; i--)
		{ut=G.FO[i];
		 if (i<G.KAO[vt-1]) {vt--;}
         if (vt>ut)
		 {if (G.KAO[ut]-G.KAO[ut-1]+G.KAO[vt]-G.KAO[vt-1]<7)
		   {u=ut;
            v=vt;
            p=G.PArray[i];
            break;
		   } else if (G.KAO[ut]-G.KAO[ut-1]+G.KAO[vt]-G.KAO[vt-1]<deg)
		   {u=ut;
            v=vt;
            p=G.PArray[i];
            deg=G.KAO[ut]-G.KAO[ut-1]+G.KAO[vt]-G.KAO[vt-1];
		   }
		 }
		}
       return pst_factor*(p*Branching(MergeVertex(G,u,v),1)+(1-p)*Branching(DeleteEdge(G,u,v),0));
	 } else {return pst_factor*ProbabilityOf5VertexGraph(G);}
	} 
}

double FactoringProbability(Graph G)
{ double p=1;
  G=ShootsDeletion(G,p);
  return p*Branching(G,0);	
}

// MPI functions

void GraphSendMPI(Graph G, int DestRank, int tag)
{	MPI_Send(&G.VertNumb, 1, MPI_INT, DestRank, tag, MPI_COMM_WORLD);
    MPI_Send(&G.EdgNumb, 1, MPI_INT, DestRank, tag, MPI_COMM_WORLD);
	MPI_Send(G.KAO, G.VertNumb+1, MPI_INT, DestRank, tag, MPI_COMM_WORLD);
    MPI_Send(G.FO, 2*G.EdgNumb, MPI_INT, DestRank, tag, MPI_COMM_WORLD);
    MPI_Send(G.PArray, 2*G.EdgNumb, MPI_DOUBLE, DestRank, tag, MPI_COMM_WORLD);
}

Graph GraphRecvMPI(int RecvRank, int tag)
{	Graph G;
    MPI_Recv(&G.VertNumb, 1, MPI_INT, RecvRank, tag, MPI_COMM_WORLD, &status);
    MPI_Recv(&G.EdgNumb, 1, MPI_INT, RecvRank, tag, MPI_COMM_WORLD, &status);
	MPI_Recv(G.KAO, G.VertNumb+1, MPI_INT, RecvRank, tag, MPI_COMM_WORLD, &status);
	MPI_Recv(G.FO, 2*G.EdgNumb, MPI_INT, RecvRank, tag, MPI_COMM_WORLD, &status);
    MPI_Recv(G.PArray, 2*G.EdgNumb, MPI_DOUBLE, RecvRank, tag, MPI_COMM_WORLD, &status);
	return G;
}

Graph GraphRecvAnySourceMPI(int tag)
{	Graph G;
    MPI_Recv(&G.VertNumb, 1, MPI_INT, MPI_ANY_SOURCE, tag, MPI_COMM_WORLD, &status);
    MPI_Recv(&G.EdgNumb, 1, MPI_INT, MPI_ANY_SOURCE, tag, MPI_COMM_WORLD, &status);
	MPI_Recv(G.KAO, G.VertNumb+1, MPI_INT, MPI_ANY_SOURCE, tag, MPI_COMM_WORLD, &status);
	MPI_Recv(G.FO, 2*G.EdgNumb, MPI_INT, MPI_ANY_SOURCE, tag, MPI_COMM_WORLD, &status);
    MPI_Recv(G.PArray, 2*G.EdgNumb, MPI_DOUBLE, MPI_ANY_SOURCE, tag, MPI_COMM_WORLD, &status);
	return G;
}

double FactoMPI(Graph G, double factor) 
{	double pst_factor,res;
        
    G=ParallelSeriesTransformation(G,pst_factor);
	factor=factor*pst_factor;
	if (G.EdgNumb>Number_of_Edges_to_Graph_Send) 
	{  // Number_of_Recursions++;//
		int vt=G.VertNumb,u,v,ut;
	    int deg=2*G.VertNumb+2;
		double p;
		for (int i=G.EdgNumb*2-1;  i>=G.KAO[1]; i--)
		{ut=G.FO[i];
		 if (i<G.KAO[vt-1]) {vt--;}
         if (vt>ut)
		 {if (G.KAO[ut]-G.KAO[ut-1]+G.KAO[vt]-G.KAO[vt-1]<7)
		   {u=ut;
            v=vt;
            p=G.PArray[i];
            break;
		   } else if (G.KAO[ut]-G.KAO[ut-1]+G.KAO[vt]-G.KAO[vt-1]<deg)
		   {u=ut;
            v=vt;
            p=G.PArray[i];
            deg=G.KAO[ut]-G.KAO[ut-1]+G.KAO[vt]-G.KAO[vt-1];
		   }
		 }
		}
		 G=DeleteEdge(G,u,v);
        double q=factor*(1-p);
		if (Connectivity(G)==false) {anti_RU=anti_RU+q; return FactoMPI(MergeVertex(G,u,v),p*factor);} 
	    else// if (Number_of_Recursions % Number_of_Recursions_to_Ask_Help==0)//
        {  int Helper_process;
		   MPI_Send(&my_rank,1, MPI_INT,0,tag0, MPI_COMM_WORLD);
           MPI_Send(&help_me, 1, MPI_DOUBLE, 0, tag0, MPI_COMM_WORLD);
		   MPI_Recv(&Helper_process, 1, MPI_INT, 0, tag0, MPI_COMM_WORLD, &status);
	       if (Helper_process>0)
		   { MPI_Send(&q, 1, MPI_DOUBLE,Helper_process,tag0, MPI_COMM_WORLD);
		     GraphSendMPI(G,Helper_process,tag0);
			 return FactoMPI(MergeVertex(G,u,v),factor*p);
		   } else  return FactoMPI(MergeVertex(G,u,v),factor*p)+FactoMPI(G,q);
	    }// else return FactoMPI(MergeVertex(G,u,v),factor*p)+FactoMPI(G,q);//
	}
	else {res=Branching(G,1); return res*factor;}
}


double Factoring_MPI_MasterProcess(Graph G, int Number_of_Processes)
{ double Result=0, Cancel_Command=10;
  int i,Proc_rank,Number_of_Branches=1;
  double Message,q=1;
  bool bol=true; 
  int Calls_Number=0;

  int FreeProcesses[Max_Number_of_Cores];

  for (i=0; i<2; i++)  
  {FreeProcesses[i]=1;}
  for (i=2; i<Number_of_Processes; i++)  
  {FreeProcesses[i]=0;}
  int Number_of_Free_Processes=Number_of_Processes-2;

  double p=1;
  G=ShootsDeletion(G,p);
  
       
  MPI_Send(&q, 1, MPI_DOUBLE,1,tag0, MPI_COMM_WORLD);
  GraphSendMPI(G,1,tag0);
  
  while (bol)
  { MPI_Recv(&Proc_rank, 1, MPI_INT, MPI_ANY_SOURCE, tag0, MPI_COMM_WORLD, &status);
    MPI_Recv(&Message, 1, MPI_DOUBLE, Proc_rank, tag0, MPI_COMM_WORLD, &status);
	if (Message==5) //   
	{if (Number_of_Free_Processes>0)
      {	  
        Number_of_Branches++;
		for (i=1; i<Number_of_Processes; i++)  
        {if (FreeProcesses[i]==0) {break;}}
        FreeProcesses[i]=1;
        Number_of_Free_Processes--;
      } else i=0;
      MPI_Send(&i, 1, MPI_INT,Proc_rank,tag0, MPI_COMM_WORLD);
	} else  //   //
	{Result=Result+Message;

	
	 Number_of_Branches--;
	 if (Number_of_Branches==0) bol=false;
	 Number_of_Free_Processes++; 
	 FreeProcesses[Proc_rank]=0; 
	}
	
  }
   
  for (i=1; i<Number_of_Processes; i++)
  {MPI_Send(&Cancel_Command, 1, MPI_DOUBLE, i, tag0, MPI_COMM_WORLD);}
  
  return p*Result; 
 
 }

void Factoring_MPI_HostProcess()
{ 
	Graph G;
	double res,q=0;
	   
	while (q<10)
	{
       MPI_Recv(&q, 1, MPI_DOUBLE, MPI_ANY_SOURCE, tag0, MPI_COMM_WORLD, &status);
	   if (q<10 )
	   {G=GraphRecvAnySourceMPI(tag0);
	   Number_of_Recursions=0;
      
	   res=FactoMPI(G,q);
	   MPI_Send(&my_rank,1, MPI_INT,0,tag0, MPI_COMM_WORLD);
       MPI_Send(&res, 1, MPI_DOUBLE, 0, tag0, MPI_COMM_WORLD);
	  		   } 
	}
}